"""Convenience wrapper around wipac-telemetry, so package can be used with/without it.

Based on https://github.com/WIPACrepo/rest-tools/blob/master/rest_tools/telemetry.py
"""

# pylint:skip-file

from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, TypeVar, cast

#
# First, try to import then implement wipac-telemetry
#
try:
    import wipac_telemetry.tracing_tools as wtt  # type: ignore[import]  # ignore for CI/CD

    evented = wtt.evented
    get_current_span = wtt.get_current_span
    inject_links_carrier = wtt.inject_links_carrier
    inject_span_carrier = wtt.inject_span_carrier
    respanned = wtt.respanned
    spanned = wtt.spanned

    CarrierRelation = wtt.CarrierRelation
    Span = wtt.Span
    SpanBehavior = wtt.SpanBehavior
    SpanKind = wtt.SpanKind
    SpanNamer = wtt.SpanNamer

    def set_current_span_attribute(key: str, value: Any) -> None:
        wtt.get_current_span().set_attribute(key, value)

    def inject_span_carrier_if_recording(carrier: Optional[Dict[str, Any]]) -> None:
        if wtt.get_current_span().is_recording():
            wtt.inject_span_carrier(carrier)


#
# Otherwise, dummy-implement every call
#
except ImportError:

    # fmt: off
    # See: https://stackoverflow.com/a/69030553
    F = TypeVar("F", bound=Callable[..., Any])

    def dummy_wrapper(*args: Any, **kwargs: Any) -> Callable[[F], F]:
        def decorator(fn: F) -> F:
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                return fn(*args, **kwargs)
            return cast(F, wrapper)
        return decorator
    # fmt:on

    evented = dummy_wrapper
    spanned = dummy_wrapper
    respanned = dummy_wrapper

    def dummy_func(*args: Any, **kwargs: Any) -> None:
        pass

    class DummyClass:
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            pass

    SpanNamer = DummyClass  # type: ignore[assignment, misc]
    Span = DummyClass  # type: ignore[assignment, misc]

    class SpanKind(Enum):  # type: ignore[no-redef]
        INTERNAL = 0
        SERVER = 1
        CLIENT = 2
        PRODUCER = 3
        CONSUMER = 4

    class SpanBehavior(Enum):  # type: ignore[no-redef]
        END_ON_EXIT = auto()
        DONT_END = auto()
        ONLY_END_ON_EXCEPTION = auto()

    class CarrierRelation(Enum):  # type: ignore[no-redef]
        SPAN_CHILD = auto()
        LINK = auto()

    set_current_span_attribute = dummy_func
    inject_span_carrier_if_recording = dummy_func

    def inject_span_carrier(*args: Any, **kwargs: Any) -> Dict[str, Any]:
        return {}

    def inject_links_carrier(*args: Any, **kwargs: Any) -> Dict[str, Any]:
        return {}

    def get_current_span() -> Span:  # type: ignore[misc,valid-type] # ignore 'valid-type' for CI/CD
        return Span()  # type: ignore[abstract]



"""Run integration tests for given broker_client, on broker_client_interface
classes.

Verify functionality that is abstracted away from the Queue class.
"""

# pylint:disable=invalid-name,too-many-public-methods,redefined-outer-name,unused-import

import copy
import itertools
import logging
from typing import List, Optional

import asyncstdlib as asl
import pytest
from mqclient.broker_client_interface import BrokerClient, Message
from mqclient.config import DEFAULT_RETRIES, DEFAULT_RETRY_DELAY, DEFAULT_TIMEOUT_MILLIS

from .utils import DATA_LIST, _log_recv, _log_send


def _log_recv_message(recv_msg: Optional[Message]) -> None:
    recv_data = None
    if recv_msg:
        recv_data = recv_msg.data
    _log_recv(f"{recv_msg} -> {recv_data}")


class PubSubBrokerClientInterface:
    """Integration test suite for broker_client_interface objects.

    Only test things that cannot be tested via the Queue class.
    """

    broker_client: BrokerClient
    timeout = 1

    @pytest.mark.asyncio
    async def test_00(self, queue_name: str, auth_token: str) -> None:
        """Sanity test."""
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        # send
        for msg in DATA_LIST:
            raw_data = Message.serialize(msg)
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive
        for i in itertools.count():
            logging.info(i)
            assert i <= len(DATA_LIST)

            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            # check received message
            if i == len(DATA_LIST):
                assert not recv_msg  # None signifies end of stream
                break

            assert recv_msg
            assert DATA_LIST[i] == recv_msg.data

            await sub.ack_message(
                recv_msg,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_10(self, queue_name: str, auth_token: str) -> None:
        """Test nacking, front-loaded sending.

        Order is not guaranteed on redelivery.
        """
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        # send
        for msg in DATA_LIST:
            raw_data = Message.serialize(msg)
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive -- nack each message, once, and anticipate its redelivery
        nacked_msgs: List[Message] = []
        redelivered_msgs: List[Message] = []
        for i in itertools.count():
            logging.info(i)
            assert i < len(DATA_LIST) * 10  # large enough but avoids inf loop

            # all messages have been acked and redelivered
            if len(redelivered_msgs) == len(DATA_LIST):
                redelivered_data = [m.data for m in redelivered_msgs]
                assert all((d in DATA_LIST) for d in redelivered_data)
                break

            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            if not recv_msg:
                logging.info("waiting...")
                continue
            assert recv_msg.data in DATA_LIST

            # message was redelivered, so ack it
            if recv_msg in nacked_msgs:
                logging.info("REDELIVERED!")
                nacked_msgs.remove(recv_msg)
                redelivered_msgs.append(recv_msg)
                await sub.ack_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
            # otherwise, nack message
            else:
                nacked_msgs.append(recv_msg)
                await sub.reject_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
                logging.info("NACK!")

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_11(self, queue_name: str, auth_token: str) -> None:
        """Test nacking, mixed sending and receiving.

        Order is not guaranteed on redelivery.
        """
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        data_to_send = copy.deepcopy(DATA_LIST)
        nacked_msgs: List[Message] = []
        redelivered_msgs: List[Message] = []
        for i in itertools.count():
            logging.info(i)
            assert i < len(DATA_LIST) * 10  # large enough but avoids inf loop

            # all messages have been acked and redelivered
            if len(redelivered_msgs) == len(DATA_LIST):
                redelivered_data = [m.data for m in redelivered_msgs]
                assert all((d in DATA_LIST) for d in redelivered_data)
                break

            # send a message
            if data_to_send:
                msg = data_to_send[0]
                raw_data = Message.serialize(msg)
                await pub.send_message(
                    raw_data,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
                _log_send(msg)
                data_to_send.remove(msg)

            # get a message
            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            if not recv_msg:
                logging.info("waiting...")
                continue
            assert recv_msg.data in DATA_LIST

            # message was redelivered, so ack it
            if recv_msg in nacked_msgs:
                logging.info("REDELIVERED!")
                nacked_msgs.remove(recv_msg)
                redelivered_msgs.append(recv_msg)
                await sub.ack_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
            # otherwise, nack message
            else:
                nacked_msgs.append(recv_msg)
                await sub.reject_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
                logging.info("NACK!")

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_20(self, queue_name: str, auth_token: str) -> None:
        """Sanity test message generator."""
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        # send
        for msg in DATA_LIST:
            raw_data = Message.serialize(msg)
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive
        last = 0
        recv_msg: Optional[Message]
        async for i, recv_msg in asl.enumerate(
            sub.message_generator(
                timeout=self.timeout,
                propagate_error=True,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
        ):
            logging.info(i)
            _log_recv_message(recv_msg)
            assert recv_msg
            assert recv_msg.data in DATA_LIST
            last = i
            await sub.ack_message(
                recv_msg,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )

        assert last == len(DATA_LIST) - 1

        await pub.close()
        await sub.close()



"""Utility data and functions."""

import logging
from typing import Any, List, Optional

import pytest
from mqclient.queue import Queue


def is_inst_name(obj: Any, name: str) -> bool:
    """Return the object's name, fully qualified with its module's name."""
    obj_name = f"{obj.__class__.__module__}.{obj.__class__.__name__}"

    return obj_name == name or obj_name.endswith("." + name)


@pytest.fixture
def queue_name() -> str:
    """Get random queue name."""
    name = Queue.make_name()
    logging.info(f"NAME :: {name}")
    return name


# Note: don't put in duplicates
DATA_LIST = [
    {"abcdefghijklmnop": ["foo", "bar", 3, 4]},
    111,
    "two",
    [1, 2, 3, 4],
    False,
    None,
]


def _log_recv(data: Any) -> Any:
    _log_data("RECV", data)
    return data


def _log_recv_multiple(data: List[Any]) -> List[Any]:
    _log_data("RECV", data, is_list=True)
    return data


def _log_send(data: Any) -> None:
    _log_data("SEND", data)


def _log_data(_type: str, data: Any, is_list: bool = False) -> None:
    if (_type == "RECV") and is_list and isinstance(data, list):
        logging.info(f"{_type} - {len(data)} :: {data}")
    else:
        logging.info(f"{_type} :: {data}")


def all_were_received(recvd: List[Any], expected: Optional[List[Any]] = None) -> bool:
    """Return True if `recvd` list is set equal to `expected`.

    If `expected` is None, use DATA_LIST.
    """
    if expected is None:  # don't override `[]`
        expected = DATA_LIST

    def log_false() -> bool:
        logging.critical(f"received ({len(recvd)}): {recvd}")
        logging.critical(
            f"expected ({len(expected) if expected is not None else 'none'}): {expected}"
        )
        return False

    # can't do set() b/c objects aren't guaranteed to be hashable

    for thing in expected:
        if thing not in recvd:
            return log_false()

    for thing in recvd:
        if thing not in expected:
            return log_false()

    if len(recvd) != len(expected):
        return log_false()

    return True



"""Utilities."""


import asyncio
import inspect
import logging
from typing import Awaitable, Callable, Optional, TypeVar, Union

from ..broker_client_interface import MQClientException

T = TypeVar("T")  # the callable/awaitable return type


def _ci_test_retry_trigger(i: int) -> None:
    pass


async def auto_retry_call(
    func: Callable[[], Union[T, Awaitable[T]]],
    retries: int,
    retry_delay: float,
    logger: logging.Logger,
    close: Optional[Callable[[], Awaitable[None]]],
    connect: Optional[Callable[[], Awaitable[None]]],
    nonretriable_conditions: Optional[Callable[[Exception], bool]],
) -> T:
    """Call `func` with auto-retries."""
    retry_delay = max(retry_delay, 0.01)
    retries = max(retries, 0)

    for i in range(retries + 1):
        try:
            _ci_test_retry_trigger(i)  # only used for testing
            ret = func()
            if inspect.isawaitable(ret):
                return await ret  # type: ignore[no-any-return]
            else:
                return ret  # type: ignore[return-value]
        except Exception as e:
            logger.error(f"Broker client action failed with {type(e).__name__}")
            logger.exception(e)
            if nonretriable_conditions and nonretriable_conditions(e):
                logger.error(
                    f"Broker client action failure ({type(e).__name__}) is not retriable"
                )
                raise
            elif i == retries:
                logger.error(
                    f"Broker client action failure ({type(e).__name__}) was final try -- raising..."
                )
                raise
            else:
                logger.warning(
                    f"Broker client action failure ({type(e).__name__}) is retriable -- trying again (attempt #{i+2})..."
                )

        # close, wait, reconnect
        if close:
            try:
                await close()  # the previous error could've been due to a closed connection
            except:  # noqa: E722
                pass
        await asyncio.sleep(retry_delay)
        if connect:
            await connect()

    # fall through -- this should not be reached in any situation
    raise MQClientException("unknown error in auto_retry_call()")



"""Package-wide config."""


DEFAULT_TIMEOUT = 60
DEFAULT_EXCEPT_ERRORS = True
DEFAULT_PREFETCH = 1
MIN_PREFETCH = 1
DEFAULT_RETRIES = 2
DEFAULT_RETRY_DELAY = 0.01

DEFAULT_TIMEOUT_MILLIS = 1000  # milliseconds


# NOTE - it's tempting to add environment variables, but since
# an application is likely to have multiple Queue instances,
# env vars are best handled on the app-side (not the API)



"""Back-end using NATS."""


import logging
import math
from typing import Any, AsyncGenerator, List, Optional, TypeVar, cast

import nats

from .. import broker_client_interface, log_msgs
from ..broker_client_interface import (
    ClosingFailedException,
    Message,
    MQClientException,
    Pub,
    RawQueue,
    Sub,
)
from ..config import DEFAULT_TIMEOUT_MILLIS
from . import utils

LOGGER = logging.getLogger("mqclient.nats")

T = TypeVar("T")  # the callable/awaitable return type


async def _anext(gen: AsyncGenerator[Any, Any], default: Any) -> Any:
    """Provide the functionality of python 3.10's `anext()`.

    https://docs.python.org/3/library/functions.html#anext
    """
    try:
        return await gen.__anext__()
    except StopAsyncIteration:
        return default


class NATS(RawQueue):
    """Base NATS wrapper, using JetStream.

    Extends:
        RawQueue
    """

    def __init__(self, endpoint: str, stream_id: str, subject: str) -> None:
        super().__init__()
        LOGGER.info(
            f"Requested MQClient for stream_id/subject '{stream_id}/{subject}' @ {endpoint}"
        )

        self.endpoint = endpoint
        self.subject = subject
        self.stream_id = stream_id

        self._nats_client: Optional[nats.aio.client.Client] = None
        self.js: Optional[nats.js.JetStreamContext] = None

        LOGGER.debug(f"Stream & Subject: {stream_id}/{self.subject}")

    async def connect(self) -> None:
        """Set up connection and channel."""
        await super().connect()
        self._nats_client = await nats.connect(self.endpoint)  # type: ignore[arg-type]
        # Create JetStream context
        self.js = self._nats_client.jetstream(timeout=DEFAULT_TIMEOUT_MILLIS // 1000)
        await self.js.add_stream(name=self.stream_id, subjects=[self.subject])

    async def close(self) -> None:
        """Close connection."""
        await super().close()
        if not self._nats_client:
            raise ClosingFailedException("No connection to close.")
        await self._nats_client.close()


class NATSPub(NATS, Pub):
    """Wrapper around PublisherClient, using JetStream.

    Extends:
        NATS
        Pub
    """

    def __init__(self, endpoint: str, stream_id: str, subject: str):
        LOGGER.debug(f"{log_msgs.INIT_PUB} ({endpoint}; {stream_id}; {subject})")
        super().__init__(endpoint, stream_id, subject)
        # NATS is pub-centric, so no extra instance needed

    async def connect(self) -> None:
        """Set up pub, then create topic and any subscriptions indicated."""
        LOGGER.debug(log_msgs.CONNECTING_PUB)
        await super().connect()
        LOGGER.debug(log_msgs.CONNECTED_PUB)

    async def close(self) -> None:
        """Close pub (no-op)."""
        LOGGER.debug(log_msgs.CLOSING_PUB)
        await super().close()
        LOGGER.debug(log_msgs.CLOSED_PUB)

    async def send_message(
        self,
        msg: bytes,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Send a message (publish)."""
        LOGGER.debug(log_msgs.SENDING_MESSAGE)
        if not self.js:
            raise MQClientException("JetStream is not connected")

        async def _send_msg():
            # use wrapper function so connection references can be updated by reconnects
            if not self.js:
                raise MQClientException("JetStream is not connected")
            return await self.js.publish(self.subject, msg)

        ack: nats.js.api.PubAck = await utils.auto_retry_call(
            func=_send_msg,
            retries=retries,
            retry_delay=retry_delay,
            close=self.close,
            connect=self.connect,
            nonretriable_conditions=None,
            logger=LOGGER,
        )
        LOGGER.debug(f"Sent Message w/ Ack: {ack}")
        LOGGER.debug(log_msgs.SENT_MESSAGE)


class NATSSub(NATS, Sub):
    """Wrapper around queue with prefetch-queue, using JetStream.

    Extends:
        NATS
        Sub
    """

    def __init__(
        self,
        endpoint: str,
        stream_id: str,
        subject: str,
        prefetch: int,
    ):
        LOGGER.debug(f"{log_msgs.INIT_SUB} ({endpoint}; {stream_id}; {subject})")
        super().__init__(endpoint, stream_id, subject)
        self._subscription: Optional[nats.js.JetStreamContext.PullSubscription] = None
        self.prefetch = prefetch

    async def connect(self) -> None:
        """Set up sub (pull subscription)."""
        LOGGER.debug(log_msgs.CONNECTING_SUB)
        await super().connect()
        if not self.js:
            raise MQClientException("JetStream is not connected.")

        self._subscription = await self.js.pull_subscribe(self.subject, "psub")
        LOGGER.debug(log_msgs.CONNECTED_SUB)

    async def close(self) -> None:
        """Close sub."""
        LOGGER.debug(log_msgs.CLOSING_SUB)
        if not self._subscription:
            raise ClosingFailedException("No sub to close.")
        await super().close()
        LOGGER.debug(log_msgs.CLOSED_SUB)

    @staticmethod
    def _to_message(  # type: ignore[override]  # noqa: F821 # pylint: disable=W0221
        msg: nats.aio.msg.Msg,  # pylint: disable=no-member
    ) -> Optional[Message]:
        """Transform NATS-Message to Message type."""
        return Message(msg.reply, msg.data)

    def _from_message(self, msg: Message) -> nats.aio.msg.Msg:
        """Transform Message instance to NATS-Message.

        Assumes the message came from this NATSSub instance.
        """
        if not self._nats_client:
            raise MQClientException("Client is not connected")

        return nats.aio.msg.Msg(
            _client=self._nats_client,
            subject=self.subject,
            reply=cast(str, msg.msg_id),  # we know this is str b/c `_to_message()`
            data=msg.data,
            headers=None,  # default
        )

    async def _get_messages(
        self,
        timeout_millis: Optional[int],
        num_messages: int,
        retries: int,
        retry_delay: float,
    ) -> List[Message]:
        """Get n messages.

        The subscriber pulls a specific number of messages. The actual
        number of messages pulled may be smaller than `num_messages`.
        """
        if not self._subscription:
            raise MQClientException("Subscriber is not connected")

        if not timeout_millis:
            timeout_millis = DEFAULT_TIMEOUT_MILLIS

        async def _get_msg():
            # use wrapper function so connection references can be updated by reconnects
            if not self._subscription:
                raise MQClientException("Subscriber is not connected")
            return await self._subscription.fetch(
                batch=num_messages,
                timeout=int(math.ceil(timeout_millis / 1000)),
            )

        try:
            nats_msgs: List[nats.aio.msg.Msg] = await utils.auto_retry_call(
                func=_get_msg,
                retries=retries,
                retry_delay=retry_delay,
                close=self.close,
                connect=self.connect,
                logger=LOGGER,
                nonretriable_conditions=lambda e: isinstance(
                    e, nats.errors.TimeoutError
                ),
            )
        except nats.errors.TimeoutError:
            LOGGER.debug(log_msgs.GETMSG_TIMEOUT_ERROR)
            return []
        if not nats_msgs:
            LOGGER.debug(log_msgs.GETMSG_NO_MESSAGE)
            return []
        msgs = []
        for recvd in nats_msgs:
            if msg := self._to_message(recvd):
                LOGGER.debug(f"{log_msgs.GETMSG_RECEIVED_MESSAGE} ({msg}).")
                msgs.append(msg)
        return msgs

    async def get_message(
        self,
        timeout_millis: Optional[int],
        retries: int,
        retry_delay: float,
    ) -> Optional[Message]:
        """Get a message."""
        LOGGER.debug(log_msgs.GETMSG_RECEIVE_MESSAGE)
        if not self._subscription:
            raise MQClientException("Subscriber is not connected.")

        try:
            msg = (
                await self._get_messages(
                    timeout_millis,
                    1,
                    retries,
                    retry_delay,
                )
            )[0]
            return msg
        except IndexError:
            return None

    async def _gen_messages(
        self,
        timeout_millis: Optional[int],
        num_messages: int,
        retries: int,
        retry_delay: float,
    ) -> AsyncGenerator[Message, None]:
        """Continuously generate messages until there are no more."""
        if not self._subscription:
            raise MQClientException("Subscriber is not connected.")

        while True:
            msgs = await self._get_messages(
                timeout_millis,
                num_messages,
                retries,
                retry_delay,
            )
            if not msgs:
                return
            for msg in msgs:
                yield msg

    async def ack_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Ack a message from the queue."""
        LOGGER.debug(log_msgs.ACKING_MESSAGE)
        if not self._subscription:
            raise MQClientException("subscriber is not connected")

        async def _ack_msg():
            # use wrapper function so connection references can be updated by reconnects
            return await self._from_message(msg).ack()

        # Acknowledges the received messages so they will not be sent again.
        await utils.auto_retry_call(
            func=_ack_msg,
            retries=retries,
            retry_delay=retry_delay,
            close=self.close,
            connect=self.connect,
            nonretriable_conditions=None,
            logger=LOGGER,
        )
        LOGGER.debug(f"{log_msgs.ACKED_MESSAGE} ({msg.msg_id!r}).")

    async def reject_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Reject (nack) a message from the queue."""
        LOGGER.debug(log_msgs.NACKING_MESSAGE)
        if not self._subscription:
            raise MQClientException("subscriber is not connected")

        async def _nack_msg():
            # use wrapper function so connection references can be updated by reconnects
            return await self._from_message(msg).nak()  # yes, it's "nak"

        await utils.auto_retry_call(
            func=_nack_msg,
            retries=retries,
            retry_delay=retry_delay,
            close=self.close,
            connect=self.connect,
            nonretriable_conditions=None,
            logger=LOGGER,
        )
        LOGGER.debug(f"{log_msgs.NACKED_MESSAGE} ({msg.msg_id!r}).")

    async def message_generator(
        self,
        timeout: int,
        propagate_error: bool,
        retries: int,
        retry_delay: float,
    ) -> AsyncGenerator[Optional[Message], None]:
        """Yield Messages.

        Generate messages with variable timeout.
        Yield `None` on `throw()`.

        Keyword Arguments:
            timeout {int} -- timeout in seconds for inactivity (default: {60})
            propagate_error {bool} -- should errors from downstream code kill the generator? (default: {True})
        """
        LOGGER.debug(log_msgs.MSGGEN_ENTERED)
        if not self._subscription:
            raise MQClientException("subscriber is not connected")

        msg = None
        try:
            gen = self._gen_messages(
                timeout * 1000,
                self.prefetch,  # prefetch = # of msgs pulled
                retries,
                retry_delay,
            )
            while True:
                # get message
                LOGGER.debug(log_msgs.MSGGEN_GET_NEW_MESSAGE)
                msg = await _anext(gen, None)
                if msg is None:
                    LOGGER.info(log_msgs.MSGGEN_NO_MESSAGE_LOOK_BACK_IN_QUEUE)
                    break

                # yield message to consumer
                try:
                    LOGGER.debug(f"{log_msgs.MSGGEN_YIELDING_MESSAGE} [{msg}]")
                    yield msg
                # consumer throws Exception...
                except Exception as e:  # pylint: disable=W0703
                    LOGGER.debug(log_msgs.MSGGEN_DOWNSTREAM_ERROR)
                    if propagate_error:
                        LOGGER.debug(log_msgs.MSGGEN_PROPAGATING_ERROR)
                        raise
                    LOGGER.warning(
                        f"{log_msgs.MSGGEN_EXCEPTED_DOWNSTREAM_ERROR} {e}.",
                        exc_info=True,
                    )
                    yield None  # hand back to consumer
                # consumer requests again, aka next()
                else:
                    pass

        # garbage collection (or explicit generator close(), or break in consumer's loop)
        except GeneratorExit:
            LOGGER.debug(log_msgs.MSGGEN_GENERATOR_EXITING)
            LOGGER.debug(log_msgs.MSGGEN_GENERATOR_EXITED)


class BrokerClient(broker_client_interface.BrokerClient):
    """NATS Pub-Sub BrokerClient Factory.

    Extends:
        BrokerClient
    """

    NAME = "nats"

    @staticmethod
    async def create_pub_queue(
        address: str,
        name: str,
        auth_token: str,
    ) -> NATSPub:
        """Create a publishing queue.

        # NOTE - `auth_token` is not used currently
        """
        if auth_token:
            LOGGER.warning("NATS broker client does not use 'auth_token'")

        q = NATSPub(  # pylint: disable=invalid-name
            address,
            name + "-stream",
            name + "-subject",
        )
        await q.connect()
        return q

    @staticmethod
    async def create_sub_queue(
        address: str,
        name: str,
        prefetch: int,
        auth_token: str,
    ) -> NATSSub:
        """Create a subscription queue.

        # NOTE - `auth_token` is not used currently
        """
        if auth_token:
            LOGGER.warning("NATS broker client does not use 'auth_token'")

        q = NATSSub(  # pylint: disable=invalid-name
            address,
            name + "-stream",
            name + "-subject",
            prefetch,
        )
        await q.connect()
        return q
